/* ************************************************************ */
/*                               esed                           */
/* A driver for the ESE line of Master Clocks, using the unit's */
/* RS-232C Output.                                              */
/*                                                              */
/*                  (C) Copyright 1996 Fred Gleason             */
/* ************************************************************ */

/* Includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/resource.h>

/* Function Prototypes */
void Detach();

/* Version String */
#define VERSION "Esed v0.0.2"

main(char argc,char *argv[])
{
  /* General Purpose Counters */
  int i;

  /* Mode Flags */
  int dDebug=0;     /* Debug mode (-d) */
  int dNoSync=0;    /* No sync mode (-n) */
  long dOffset=0;   /* Offset value in hours (-o) */
  int dSyncSec=47;  /* Second to sync time on (-s) */
  int dOptionValid; /* Valid option found */

  /* Stream Stuff */
  FILE *hTty;       /* Handle to tty stream */

  /* Parser Variables */
  char c;           /* Tty character buffer */
  int iState;       /* Parser state */
  struct tm tmEseTime;  /* Ese time */
  time_t ttSystemTime;  /* System time */
  char sAccum[5];       /* String accumulator */

  /* Check for -v option */
  if(argc>1) {
    for(i=1;i<argc;i++) {
      if(strcmp(argv[i],"-v")==0) {   /* Version message */
	printf("\n");
	printf(VERSION);
	printf("\n\n");
	exit(0);
      }
    }
  }
  /* Check for other options */
  if(argc>2) {
    for(i=1;i<argc-1;i++) {
      dOptionValid=0;
      if(strcmp(argv[i],"-d")==0) {   /* Debug mode */
	dOptionValid=1;
	dDebug=1;
      }
      if(strcmp(argv[i],"-n")==0) {   /* No syncronize mode */
	dOptionValid=1;
	dNoSync=1;
      }
      if(strcmp(argv[i],"-s")==0) {   /* Set sync second */
	dOptionValid=1;
	if(i+2>=argc) {     /* missing argument */
	  printf("esed: missing argument\n");
	  exit(1);
	}
	i++;
	if(sscanf(argv[i],"%d",&dSyncSec)!=1) {
	  printf("esed: invalid argument\n");
	  exit(1);
	}
	if(dSyncSec<0 || dSyncSec>59) {
	  printf("esed: invalid argument\n");
	  exit(1);
	}
      }
      if(strcmp(argv[i],"-o")==0) {   /* Offset value */
	dOptionValid=1;
	if(i+2>=argc) {     /* missing argument */
	  printf("esed: missing argument\n");
	  exit(1);
	}
	i++;
	if(sscanf(argv[i],"%ld",&dOffset)!=1) {
	  printf("esed: invalid argument\n");
	  exit(1);
	}
      }
      if(dOptionValid==0) {
	printf("esed: invalid argument\n");
	exit(1);
      }  
    }
  }

  /* Open the tty */
  if((hTty=fopen(argv[argc-1],"r"))==NULL) {
    printf("esed: can't open %s\n",argv[argc-1]);
    exit(1);
  }

  /* We're time critical by definition, so get a good priority */
/*  setpriority(PRIO_PROCESS,0,-15);  */

  /* Everything seems OK, so let's detach */
  if(dDebug==0) {
    Detach();
  }

  /* Daylight savings time stuff */
  ttSystemTime=time(NULL);
  tmEseTime=*localtime(&ttSystemTime);

  /* OK, here's the main loop */
  while(1==1) {
    fscanf(hTty,"%c",&c);
    switch(iState) {

    case 0:    /* First digit of month */
      if(c>='0' && c<='1') {
	sAccum[0]=c;
	iState=1;
      }
      else {
	iState=0;
      }
    break;

    case 1:    /* Second digit of month */
      if(c>='0' && c<='9') {
	sAccum[1]=c;
	sAccum[2]=0;
	sscanf(sAccum,"%d",&tmEseTime.tm_mon);
	tmEseTime.tm_mon--;
	iState=2;
      }
      else {
	iState=0;
      }
    break;
    
    case 2:    /* Seperator */
      if(c=='-') {
	iState=3;
      }
      else {
	iState=0;
      }
    break;

    case 3:    /* First digit of day */
      if(c>='0' && c<='3') {
	sAccum[0]=c;
	iState=4;
      }
      else {
	iState=0;
      }
    break;

    case 4:    /* Second digit of day */
      if(c>='0' && c<='9') {
	sAccum[1]=c;
	sAccum[2]=0;
	sscanf(sAccum,"%d",&tmEseTime.tm_mday);
	iState=5;
      }
      else {
	iState=0;
      }
    break;

    case 5:    /* Seperator */
      if(c=='-') {
	iState=6;
      }
      else {
	iState=0;
      }
    break;

    case 6:    /* First digit of year */
      if(c>='0' && c<='9') {
	sAccum[0]=c;
	iState=7;
      }
      else {
	iState=0;
      }
    break;

    case 7:    /* Second digit of year */
      if(c>='0' && c<='9') {
	sAccum[1]=c;
	sAccum[2]=0;
	sscanf(sAccum,"%d",&tmEseTime.tm_year);
	if(tmEseTime.tm_year<96) {
	  tmEseTime.tm_year+=100;   /* Welcome to the 21st century! */
	}
	iState=8;
      }
      else {
	iState=0;
      }
    break;

    case 8:    /* Space */
      if(c==' ') {
	iState=9;
      }
      else {
	iState=0;
      }
    break;

    case 9:    /* Space */
      if(c==' ') {
	iState=10;
      }
      else {
	iState=0;
      }
    break;

    case 10:    /* First digit of julian day */
      if(c>='0' && c<='3') {
	sAccum[0]=c;
	iState=11;
      }
      else {
	iState=0;
      }
    break;

    case 11:    /* Second digit of julian day */
      if(c>='0' && c<='9') {
	sAccum[1]=c;
	iState=12;
      }
      else {
	iState=0;
      }
    break;

    case 12:    /* Third digit of julian day */
      if(c>='0' && c<='9') {
	sAccum[2]=c;
	sAccum[3]=0;
	sscanf(sAccum,"%d",&tmEseTime.tm_yday);
	tmEseTime.tm_yday--;
	iState=13;
      }
      else {
	iState=0;
      }
    break;

    case 13:    /* Seperator */
      if(c==':') {
	iState=14;
      }
      else {
	iState=0;
      }
    break;

    case 14:    /* First digit of hour */
      if(c>='0' && c<='2') {
	sAccum[0]=c;
	iState=15;
      }
      else {
	iState=0;
      }
    break;

    case 15:    /* Second digit of hour */
      if(c>='0' && c<='9') {
	sAccum[1]=c;
	sAccum[2]=0;
	sscanf(sAccum,"%d",&tmEseTime.tm_hour);
	iState=16;
      }
      else {
	iState=0;
      }
    break;

    case 16:    /* Seperator */
      if(c==':') {
	iState=17;
      }
      else {
	iState=0;
      }
    break;

    case 17:    /* First digit of minute */
      if(c>='0' && c<='5') {
	sAccum[0]=c;
	iState=18;
      }
      else {
	iState=0;
      }
    break;

    case 18:    /* Second digit of minute */
      if(c>='0' && c<='9') {
	sAccum[1]=c;
	sAccum[2]=0;
	sscanf(sAccum,"%d",&tmEseTime.tm_min);
	iState=19;
      }
      else {
	iState=0;
      }
    break;

    case 19:    /* Seperator */
      if(c==':') {
	iState=20;
      }
      else {
	iState=0;
      }
    break;

    case 20:    /* First digit of seconds */
      if(c>='0' && c<='6') {
	sAccum[0]=c;
	iState=21;
      }
      else {
	iState=0;
      }
    break;

    case 21:    /* Second digit of seconds */
      if(c>='0' && c<='9') {
	sAccum[1]=c;
	sAccum[2]=0;
	sscanf(sAccum,"%d",&tmEseTime.tm_sec);
	ttSystemTime=mktime(&tmEseTime);
	ttSystemTime+=(3600*dOffset);
	tmEseTime=*localtime(&ttSystemTime);
	iState=22;
      }
      else {
	iState=0;
      }
    break;

    case 22:    /* Sync character */
      if(c==0x0A) {
	if(tmEseTime.tm_sec==dSyncSec) {
	  if(dNoSync==0) {
	    stime(&ttSystemTime);
	    if(dDebug==1) {
	      printf("Sync!\n");
	    }
	  }
	}
	if(dDebug==1) {
	  /* print month */
	  if(tmEseTime.tm_mon<9) {
	    printf("0%d-",tmEseTime.tm_mon+1);
	  }
	  else {
	    printf("%d-",tmEseTime.tm_mon+1);
	  }
	  
	  /* print day */
	  if(tmEseTime.tm_mday<10) {
	    printf("0%d-",tmEseTime.tm_mday);
	  }
	  else {
	    printf("%d-",tmEseTime.tm_mday);
	  }
	  
	  /* print year */
	  if(tmEseTime.tm_year<100) {
	    printf("%d  ",tmEseTime.tm_year);
	  }
	  else {
	    printf("%d  ",tmEseTime.tm_year-100);
	  }
	  
	  /* print julian day */
	  if(tmEseTime.tm_yday<10) {
	    printf("00%d:",tmEseTime.tm_yday+1);
	  }
	  if(tmEseTime.tm_yday>9 && tmEseTime.tm_yday<100) {
	    printf("0%d:",tmEseTime.tm_yday);
	  }
	  if(tmEseTime.tm_yday>=100) {
	    printf("%d:",tmEseTime.tm_yday);
	  }
	  
	  /* print hours */
	  if(tmEseTime.tm_hour<10) {
	    printf("0%d:",tmEseTime.tm_hour);
	  }
	  else {
	    printf("%d:",tmEseTime.tm_hour);
	  }

	  /* print minutes */
	  if(tmEseTime.tm_min<10) {
	    printf("0%d:",tmEseTime.tm_min);
	  }
	  else {
	    printf("%d:",tmEseTime.tm_min);
	  }
	  
	  /* print seconds */
	  if(tmEseTime.tm_sec<10) {
	    printf("0%d\n",tmEseTime.tm_sec);
	  }
	  else {
	    printf("%d\n",tmEseTime.tm_sec);
	  }
	  
	}
      }
      iState=0;
    break;

    default:
      iState=0;
    break;
    }
  }
}



void Detach() 
{
  int i;
  
  chdir("/");
  if((i = fork()) > 0)
    exit(0);
  else if(i == -1) {
    printf("esed: unable to fork new process\n");
    perror("fork");
    exit(1);
  }
  if(setsid() == -1) {
    printf("esed: setsid failed\n");
    perror("setsid");
    exit(1);
  }    
}












